/*
 * Copyright (c) 1995, 1996, 1997, 1998 Kungliga Tekniska Hgskolan
 * (Royal Institute of Technology, Stockholm, Sweden).
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Kungliga Tekniska
 *      Hgskolan and its contributors.
 * 
 * 4. Neither the name of the Institute nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * Test to talk with FS
 */

#include "arla_local.h"
#include <sl.h>

RCSID("$Id: arla.c,v 1.45 1998/02/23 07:01:52 assar Exp $") ;

#define AFSNAMEMAX 256

static CredCacheEntry *tmpce;

static int arla_chdir(int, char **);
static int arla_ls(int, char **);
static int arla_cat(int, char **);
static int help(int, char **);
static int arla_quit(int, char **);
static int arla_conn_status(int, char **);
static int arla_vol_status(int, char **);
static int arla_cred_status(int, char **);
static int arla_fcache_status(int, char **);
static int arla_sysname(int, char**);
static int arla_rx_status(int argc, char **argv);

static SL_cmd cmds[] = {
    {"chdir", arla_chdir, "chdir directory"},
    {"cd"},
    {"ls",    arla_ls, "ls"},
    {"cat",   arla_cat, "cat file"},
    {"help",  help, "help"},
    {"?"},
    {"conn-status", arla_conn_status, "connection status"},
    {"vol-status", arla_vol_status, "volume cache status"},
    {"cred-status", arla_cred_status, "credentials status"},
    {"fcache-status", arla_fcache_status, "file cache status"},
    {"rx-status", arla_rx_status, "rx connection status"},
    {"quit", arla_quit, "quit"},
    {"exit"},
    {"sysname", arla_sysname, "sysname"},
    { NULL }
};

static void
print_dir (VenusFid *fid, const char *name, void *v)
{
     printf("(%d, %d, %d, %d): %s\n", fid->Cell,
	    fid->fid.Volume,
	    fid->fid.Vnode,
	    fid->fid.Unique, name);
}

static VenusFid
newwalk (VenusFid startdir, char *fname)
{
     VenusFid cwd = startdir;
     char *slash;
     char *comp;
     VenusFid file;
     Result ret;

     do {
	 comp = fname;
	 slash = strchr (fname, '/');
	 if (slash) {
	     *slash = '\0';
	     fname = slash + 1;
	 }
	 ret = cm_lookup (cwd, comp, &file, &tmpce);
	 if (ret.res) {
	     ARLADEB (ADEBMISC, ("Error in lookup\n"));
	     return cwd;
	 }
	 cwd = file;
     } while(slash != NULL && *comp);
     return cwd;
}

static VenusFid cwd;

static int
arla_quit (int argc, char **argv)
{
    printf("Thank you for using arla\n");
    return 1;
}

static int
arla_chdir (int argc, char **argv)
{
    if (argc != 2) {
	printf ("usage: %s dir\n", argv[0]);
	return 0;
    }

    cwd = newwalk (cwd, argv[1]);
    return 0;
}

static int
arla_ls (int argc, char **argv)
{
    int error;

    error = adir_readdir (cwd, print_dir, NULL, tmpce);
    if (error) {
	printf ("adir_readdir failed: %d\n", error);
    }
    return 0;
}

static int
arla_sysname (int argc, char **argv)
{
    switch (argc) {
    case 1:
	printf("sysname: %s\n", arlasysname);
	break;
    case 2:
	strncpy(arlasysname, argv[1], SYSNAMEMAXLEN);
	arlasysname[SYSNAMEMAXLEN-1] = '\0';
	printf("setting sysname to: %s\n", arlasysname);
	break;
    default:
	printf("syntax: sysname <sysname>\n");
	break;
    }
    return 0;
}



static int
arla_cat (int argc, char **argv)
{
    VenusFid fid;
    int fd;
    char buf[8192];
    int ret;
    FCacheEntry *e;
    
    if (argc != 2) {
	printf ("usage: %s file\n", argv[0]);
	return 0;
    }
    fid = newwalk (cwd, argv[1]);

    fcache_get (&e, fid, tmpce);
    fcache_get_data (e, tmpce);

    fd = fcache_open_file (e, O_RDONLY, 0);

    if (fd < 0) {
	printf ("fcache_open_file failed: %d\n", ret);
	return 0;
    }
    while ((ret = read (fd, buf, sizeof(buf))) > 0) {
	write (STDOUT_FILENO, buf, ret);
    }
    close (fd);
    return 0;
}

static int
help (int argc, char **argv)
{
    sl_help(cmds, argc, argv);
    return 0;
}

static int
arla_conn_status (int argc, char **argv)
{
    conn_status (stderr);
    return 0;
}

static int
arla_vol_status (int argc, char **argv)
{
    volcache_status (stderr);
    return 0;
}

static int
arla_cred_status (int argc, char **argv)
{
    cred_status (stderr);
    return 0;
}

static int
arla_fcache_status (int argc, char **argv)
{
    fcache_status (stderr);
    return 0;
}

static int
arla_rx_status(int argc, char **argv)
{
    rx_PrintStats(stderr);
    return 0;
}


static void
initrx (void)
{
     int error;

     error = rx_Init (htons(4711));
     if (error) {
	  ARLADEB (ADEBMISC, ("Cannot rx_init. code = %d\n", error));
	  exit (1);
     }
}

#ifdef KERBEROS

static int
get_cred(const char *princ, const char *inst, const char *krealm, 
         CREDENTIALS *c)
{
  KTEXT_ST foo;
  int k_errno;

  k_errno = krb_get_cred((char*)princ, (char*)inst, (char*)krealm, c);

  if(k_errno != KSUCCESS) {
    k_errno = krb_mk_req(&foo, (char*)princ, (char*)inst, (char*)krealm, 0);
    if (k_errno == KSUCCESS)
      k_errno = krb_get_cred((char*)princ, (char*)inst, (char*)krealm, c);
  }
  return k_errno;
}


#endif /* KERBEROS */

#define KERNEL_STACKSIZE (16*1024)

/*
 *
 */

static void
sigint (int foo)
{
    exit (0);
}

struct conf_param {
    const char *name;
    unsigned *val;
    unsigned default_val;
};

/*
 * Reads in a configuration file, and sets some defaults.
 */

static void
read_conffile(char *fname,
	      struct conf_param *params)
{
    FILE *fp;
    char buf[256];
    int lineno;
    struct conf_param *p;

    for (p = params; p->name; ++p)
	*(p->val) = p->default_val;

    ARLADEB(ADEBINIT, ("read_conffile: %s\n", fname));

    fp = fopen(fname, "r");
    if (fp == NULL) {
	ARLADEB(ADEBINIT, ("read_conffile: no configuration file\n"));
	return;
    }

    lineno = 0;

    while (fgets (buf, sizeof(buf), fp) != NULL) {
	struct conf_param *partial_param = NULL;
	int partial_match = 0;
	char *save = NULL;
	char *n;
	char *v;
	unsigned val;
	char *endptr;

	++lineno;
	if (buf[strlen(buf) - 1] == '\n')
	    buf[strlen(buf) - 1] = '\0';
	if (buf[0] == '#')
	    continue;

	n = strtok_r (buf, " \t", &save);
	if (n == NULL) {
	    fprintf (stderr, "%s:%d: no parameter?\n", fname, lineno);
	    continue;
	}

	v = strtok_r (NULL, " \t", &save);
	if (v == NULL) {
	    fprintf (stderr, "%s:%d: no value?\n", fname, lineno);
	    continue;
	}

	val = strtol(v, &endptr, 0);
	if (endptr == v) {
	    fprintf (stderr, "%s:%d: bad value `%s'\n",
		     fname, lineno, v);
	}

	for (p = params; p->name; ++p) {
	    if (strcmp(n, p->name) == 0) {
		partial_match = 1;
		partial_param = p;
		break;
	    } else if(strncmp(n, p->name, strlen(n)) == 0) {
		++partial_match;
		partial_param = p;
	    }
	}
	if (partial_match == 1)
	    *(partial_param->val) = val;
	else if (partial_match == 0)
	    fprintf (stderr, "%s:%d: unknown parameter `%s'\n",
		     fname, lineno, n);
	else
	    fprintf (stderr, "%s:%d: ambiguous parameter `%s'\n",
		     fname, lineno, n);
    }
    fclose(fp);
}

static unsigned low_vnodes, high_vnodes, low_bytes, high_bytes;
static unsigned numcreds, numconns, numvols;

static struct conf_param conf_params[] = {
    {"low_vnodes",		&low_vnodes,	ARLA_LOW_VNODES},
    {"high_vnodes",		&high_vnodes,	ARLA_HIGH_VNODES},
    {"low_bytes",		&low_bytes,	ARLA_LOW_BYTES},
    {"high_bytes",		&high_bytes,	ARLA_HIGH_BYTES},
    {"numcreds",		&numcreds,	ARLA_NUMCREDS},
    {"numconns",		&numconns,	ARLA_NUMCREDS},
    {"numvols",			&numvols,	ARLA_NUMVOLS},
    {NULL,			NULL,		0}};

int
main (int argc, char **argv)
{
    char *conf_file;
    PROCESS kernelpid;
    int error;

#ifdef KERBEROS
    krbstruct krbdata;
#endif

    signal (SIGINT, sigint);

    umask (S_IRWXG|S_IRWXO); /* 077 */
     
    if (argc < 2) {
	fprintf(stderr,
		"Usage: %s ({device} | -t) [conffile]\n"
		"\tGive name of xfs-device to start in daemon mode.\n"
		"\tOr -t to run in test command-line version.\n"
		"\tconffile is the name of the configuration file\n"
		"\t"ARLACONFFILE" is the default\n",
		argv[0]);
	return 1;
    }

    if (argc < 3)
	conf_file = ARLACONFFILE;
    else
	conf_file = argv[2];

    read_conffile(conf_file, conf_params);

    ARLADEB(ADEBINIT, ("Arlad booting sequence:\n"));
    ARLADEB(ADEBINIT, ("initports\n"));
    initports ();
    ARLADEB(ADEBINIT, ("conn_init numconns = %u\n", numconns));
    conn_init (numconns);
    ARLADEB(ADEBINIT, ("cellcache\n"));
    cell_init ();
    ARLADEB(ADEBINIT, ("volcache numvols = %u\n", numvols));
    volcache_init (numvols);
    ARLADEB(ADEBINIT, ("rx\n"));
    initrx ();
    ARLADEB(ADEBINIT, ("credcache numcreds = %u\n", numcreds));
    cred_init (numcreds);
    ARLADEB(ADEBINIT, ("fcache low_vnodes = %u, high_vnodes = %u\n"
		       "low_bytes = %u, high_bytes = %u\n",
		       low_vnodes, high_vnodes,
		       low_bytes, high_bytes));
    fcache_init (ARLACACHEDIR, low_vnodes, high_vnodes,
		 low_bytes, high_bytes);
    ARLADEB(ADEBINIT, ("cmcb\n"));
    cmcb_init ();
    ARLADEB(ADEBINIT, ("\n"));

    if (strcmp (argv[1], "-t") == 0) {
#ifdef KERBEROS
	{
	    char *realm = strdup(cell_getthiscell ());

	    strupr(realm);
	  
	    if (get_cred("afs","",realm, &krbdata.c)) {
		ARLADEB(ADEBINIT, ("error getting ticket for %s\n",
				   realm));
	    } else if (cred_add_krb4(getuid(), &krbdata.c) == NULL) {
		ARLADEB(ADEBINIT, ("Could not insert tokens to arla\n"));
	    }
	    free (realm);
	}
#endif

	tmpce = cred_get (0, getuid(), CRED_ANY);
	 
	ARLADEB(ADEBINIT, ("Getting root...\n"));
	error = getroot (&cwd, tmpce);
	if (error) {
	    ARLADEB (ADEBINIT, ("getroot failed: %d\n", error));
	    exit (1);
	}
	ARLADEB(ADEBINIT, ("arla loop started\n"));
	sl_loop(cmds, "arla> ");
    } else {
	xfs_message_init ();

	if (LWP_CreateProcess (kernel_interface, KERNEL_STACKSIZE, 1,
			       argv[1], "Kernel-interface", &kernelpid))
	    ARLADEB (ADEBINIT, ("Cannot create kernel-interface process\n"));
	  
	LWP_WaitProcess ((char *)main);
	abort ();
    }
    return 0;
}
