/*
 * Copyright (c) 1995, 1996, 1997, 1998 Kungliga Tekniska Hgskolan
 * (Royal Institute of Technology, Stockholm, Sweden).
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Kungliga Tekniska
 *      Hgskolan and its contributors.
 * 
 * 4. Neither the name of the Institute nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * Cell information
 */

#include "arla_local.h"
RCSID("$Id: cell.c,v 1.15 1998/02/23 04:29:28 lha Exp $");

#define CELLFILENAME "CellServDB"
#define THISCELLFILENAME "ThisCell"

#define CELLCACHESIZE 499

static Hashtab *cellnamehtab, *cellnumhtab;

static char *thiscell;

/*
 * Functions for handling cell entries.
 */

static int
cellnamecmp (void *a, void *b)
{
     CellEntry *c1 = (CellEntry *)a;
     CellEntry *c2 = (CellEntry *)b;

     return strcmp (c1->name, c2->name);
}

static unsigned
cellnamehash (void *a)
{
     CellEntry *c = (CellEntry *)a;

     return hashadd (c->name);
}

static int
cellnumcmp (void *a, void *b)
{
     CellEntry *c1 = (CellEntry *)a;
     CellEntry *c2 = (CellEntry *)b;

     return c1->id != c2->id;
}

static unsigned
cellnumhash (void *a)
{
     CellEntry *c = (CellEntry *)a;

     return c->id;
}

/*
 * Record this cell in the hashtable
 */

static void
recordcell (CellEntry *c, int dbnum, DbServerEntry *dbservs)
{
     if(dbnum == 0)
	  ARLADEB (ADEBMISC, ("No db-servers for cell %s\n", c->name));

     c->ndbservs = dbnum;
     c->dbservs  = (DbServerEntry*)malloc (dbnum 
					   * sizeof(DbServerEntry));
     memcpy (c->dbservs, dbservs, dbnum * sizeof (DbServerEntry));
							 
     hashtabadd (cellnamehtab, c);
     hashtabadd (cellnumhtab, c);
}

/*
 * New cell from cellserver-database file
 */

static void
newcell (char *line, CellEntry** c, int *dbnum)
{
     static long cellno = 0;
     char name[256];
     char *hash;

     *c = (CellEntry *)malloc (sizeof (**c));
     sscanf (line, ">%s", name);
     (*c)->name = strdup (name);
     if (strcmp (name, cell_getthiscell ()) == 0)
	  (*c)->id = 0;
     else
	  (*c)->id = ++cellno;
     hash = strchr (line, '#');
     if (hash != NULL)
	  (*c)->expl = strdup (hash+1);
     else
	  (*c)->expl = NULL;
     *dbnum = 0;
}

/*
 * Read one line of database information.
 */

static int
readdb (char *line, CellEntry* c, int *dbnum, int maxdbs,
	DbServerEntry *dbservs, int lineno)
{
     struct in_addr numaddr;
     char *hostname;

     if (*dbnum >= maxdbs) {
	  ARLADEB (ADEBMISC, ("Too many database servers for "
		   "cell %s. Ignoring\n", c->name));
	  return -1;
     }
     if ( (hostname = strchr (line, '#')) == NULL
	 || (numaddr.s_addr = inet_addr (line)) == -1 ) {
	  ARLADEB (ADEBMISC, ("Syntax error at line %d in %s\n",
		   lineno, CELLFILENAME));
	  return -1;
     } else {
	  ++hostname;
	  dbservs[*dbnum].name = strdup (hostname);
	  dbservs[*dbnum].addr = numaddr;
	  ++(*dbnum);
     }
     return 0;
}

/*
 * Read the information from the cell-server db.
 */

static void
readcellservdb (const char *filename)
{
     FILE *f;
     char line[256];
     CellEntry *c = NULL;
     int lineno = 0;
     int dbnum;
     DbServerEntry dbservs[256];

     f = fopen (filename, "r");
     if (f == NULL) {
	  ARLADEB (ADEBMISC, ("Cannot read cell information from %s\n", 
		   filename));
	  exit(1);
     }

     while (fgets (line, sizeof (line), f)) {
	  ++lineno;
	  if (line[strlen(line) - 1 ] != '\n')
	       ARLADEB (ADEBMISC, ("Too long line at line %d in %s\n",
			lineno, filename));
	  else
	       line[strlen(line) - 1] = '\0';
	  if (*line == '#' || *line == '\0')
	       continue;
	  if (*line == '>') {
	       if (c)
		    recordcell (c, dbnum, dbservs);
	       newcell (line, &c, &dbnum);
	  } else {
	       if (readdb(line, c, &dbnum, sizeof (dbservs) /
			  sizeof(*dbservs),
			  dbservs, lineno))
		    continue;
	  }
     }
     if (c)
	  recordcell (c, dbnum, dbservs);
     fclose (f);
}

/*
 * Read name of this cell.
 */

static void
readthiscell (const char *filename)
{
     FILE *f;
     char cell[256];

     f = fopen (filename, "r");
     if (f == NULL) {
	 fprintf(stderr, "Cannot open file %s.\n"
		 "I don't know where I am\n",
		 filename);
	 exit (1);
     }

     if (fgets (cell, sizeof cell, f) == NULL) {
	  fprintf(stderr, "Cannot read cellname from %s\n",
		  filename);
	  exit(1);
     }
     if (cell[strlen(cell) - 1] == '\n')
	 cell[strlen(cell) - 1] = '\0';
     thiscell = strdup (cell);
     fclose (f);
}

/*
 * Initialize the cache of cell information.
 */

void
cell_init (void)
{
     cellnamehtab = hashtabnew (CELLCACHESIZE, cellnamecmp, cellnamehash);
     cellnumhtab  = hashtabnew (CELLCACHESIZE, cellnumcmp,  cellnumhash);
     readthiscell (SYSCONFDIR "/" THISCELLFILENAME);
     readcellservdb (SYSCONFDIR "/" CELLFILENAME);
}

/*
 * Find a DB server to talk to for a given cell.
 */

struct in_addr
cell_finddbserver (int32_t cell)
{
     CellEntry key;
     CellEntry *data;
     struct in_addr addr;

     key.id = cell;
     data = hashtabsearch (cellnumhtab, &key);
     if (data == NULL) {
	  ARLADEB (ADEBMISC, ("Cannot find cell %d\n", cell));
	  return addr;
     }
     if (data->ndbservs == 0) {
	  ARLADEB (ADEBMISC, ("No DB servers for cell %d\n", cell));
	  return addr;
     }
     if (ipgetaddr (data->dbservs[0].name, &addr))
	  return addr;
     else
	  return data->dbservs[0].addr;
}

/*
 * Return the ID given the name for a cell.
 * -1 if the cell does not exist.
 */

int32_t
cell_name2num (char *cell)
{
     CellEntry key, *data;

     key.name = cell;
     data = hashtabsearch (cellnamehtab, (void *)&key);
     if (data)
	  return data->id;
     else
	  return -1;
}

/*
 * Return the name given the ID or NULL if the cell doesn't exist.
 */

const char *
cell_num2name (int32_t cell)
{
    CellEntry key, *data;

    key.id = cell;
    data = hashtabsearch (cellnumhtab, (void *)&key);
    if (data)
	return data->name;
    else
	return NULL;
}

/*
 * Return name of the cell of the cache manager.
 */

const char *
cell_getthiscell (void)
{
     return thiscell;
}
