/*
 * Copyright (c) 1995, 1996, 1997, 1998 Kungliga Tekniska Hgskolan
 * (Royal Institute of Technology, Stockholm, Sweden).
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Kungliga Tekniska
 *      Hgskolan and its contributors.
 * 
 * 4. Neither the name of the Institute nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "arla_local.h"
RCSID("$Id: messages.c,v 1.28 1998/02/23 06:08:57 lha Exp $");

#include <xfs_message.h>

#ifdef KERBEROS
#include <des.h>
#include <krb.h>
#endif /* KERBEROS */
#ifdef HAVE_SYS_IOCTL_H
#include <sys/ioctl.h>
#endif
#include <kafs.h>

#include "messages.h"

static xfs_message_function rcvfuncs[] = {
NULL,						/* version */
(xfs_message_function)xfs_message_wakeup,	/* wakeup */
(xfs_message_function)xfs_message_getroot,	/* getroot */
NULL,						/* installroot */
(xfs_message_function)xfs_message_getnode, 	/* getnode */
NULL,						/* installnode */
(xfs_message_function)xfs_message_getattr,	/* getattr */
NULL,						/* installattr */
(xfs_message_function)xfs_message_getdata,	/* getdata */
NULL,						/* installdata */
(xfs_message_function)xfs_message_inactivenode,	/* inactivenode */
NULL,						/* invalidnode */
(xfs_message_function)xfs_message_getdata,	/* open */
(xfs_message_function)xfs_message_putdata,      /* put_data */
(xfs_message_function)xfs_message_putattr,      /* put attr */
(xfs_message_function)xfs_message_create,       /* create */
(xfs_message_function)xfs_message_mkdir,	/* mkdir */
(xfs_message_function)xfs_message_link,		/* link */
(xfs_message_function)xfs_message_symlink,      /* symlink */
(xfs_message_function)xfs_message_remove,	/* remove */
(xfs_message_function)xfs_message_rmdir,	/* rmdir */
(xfs_message_function)xfs_message_rename,	/* rename */
(xfs_message_function)xfs_message_pioctl,	/* pioctl */
NULL	                                        /* wakeup_data */
};

static int nfuncs = sizeof (rcvfuncs) / sizeof (*rcvfuncs);

static u_int *seqnums;

static List *sleepers;

char *rcvfuncs_name[] = 
{
  "version",
  "wakeup",
  "getroot",
  "installroot",
  "getnode",
  "installnode",
  "getattr",
  "installattr",
  "getdata",
  "installdata",
  "inactivenode",
  "invalidnode",
  "open",
  "put_data",
  "put_attr",
  "create",
  "mkdir",
  "link",
  "symlink",
  "remove",
  "rmdir",
  "rename",
  "pioctl",
  "wakeup_data",
} ;

long
afsfid2inode (VenusFid *fid)
{
     return ((fid->fid.Volume & 0x7FFF) << 16 | (fid->fid.Vnode & 0xFFFF));
}

void
afsstatus2xfs_attr (AFSFetchStatus *status,
		    VenusFid *fid,
		    struct xfs_attr *attr)
{
     attr->valid = XA_V_NONE;
     switch (status->FileType) {
	  case TYPE_FILE :
	       XA_SET_MODE(attr, S_IFREG);
	       XA_SET_TYPE(attr, XFS_FILE_REG);
	       break;
	  case TYPE_DIR :
	       XA_SET_MODE(attr, S_IFDIR);
	       XA_SET_TYPE(attr, XFS_FILE_DIR);
	       break;
	  case TYPE_LINK :
	       XA_SET_MODE(attr, S_IFLNK);
	       XA_SET_TYPE(attr, XFS_FILE_LNK);
	       break;
	  default :
	       ARLADEB (ADEBMISC, ("afsstatus2xfs_attr: default\n"));
	       abort ();
     }
     XA_SET_NLINK(attr, status->LinkCount);
     XA_SET_SIZE(attr, status->Length);
     XA_SET_UID(attr,status->Owner);
     XA_SET_GID(attr, status->Group);
     attr->xa_mode  |= status->UnixModeBits;
     XA_SET_ATIME(attr, status->ServerModTime);
     XA_SET_MTIME(attr, status->ServerModTime);
     XA_SET_CTIME(attr, status->ServerModTime);
     XA_SET_FILEID(attr, afsfid2inode(fid));
}

int
xfs_attr2afsstorestatus(struct xfs_attr *xa,
			AFSStoreStatus *storestatus)
{
    int mask = 0;

    if (XA_VALID_MODE(xa)) {
	storestatus->UnixModeBits = xa->xa_mode;
	mask |= SS_MODEBITS;
    }
    if (XA_VALID_UID(xa)) {
	storestatus->Owner = xa->xa_uid;
	mask |= SS_OWNER;
    }
    if (XA_VALID_GID(xa)) {
	storestatus->Group = xa->xa_gid;
	mask |= SS_GROUP;
    }
    if (XA_VALID_MTIME(xa)) {
	storestatus->ClientModTime = xa->xa_mtime;
	mask |= SS_MODTIME;
    }
    storestatus->Mask = mask ;   

    /* SS_SegSize */
    storestatus->SegSize = 0 ;
    return 0;
}

void
xfs_message_init (void)
{
     unsigned i;

     seqnums = (u_int *)malloc (sizeof (*seqnums) * getdtablesize ());
     if (seqnums == NULL)
	 errx (1, "xfs_message_init: malloc failed");
     for (i = 0; i < getdtablesize (); ++i)
	  seqnums[i] = 0;
     sleepers = listnew ();
     if (sleepers == NULL)
	 errx (1, "xfs_message_init: listnew failed");
}

int
xfs_message_receive (int fd, struct xfs_message_header *h, u_int size)
{
     if (h->opcode >= nfuncs || rcvfuncs[h->opcode] == NULL ) {
	  ARLADEB (ADEBMISC, ("Bad message opcode = %u\n", h->opcode));
	  return -1;
     }

     ARLADEB (ADEBMISC, ("Rec message: opcode = %u (%s), size = %u\n",
	      h->opcode, rcvfuncs_name[h->opcode], h->size));

     return (*rcvfuncs[h->opcode])(fd, h, size);
}

int
xfs_message_send (int fd, struct xfs_message_header *h, u_int size)
{
     int res;

     h->size = size;
     h->sequence_num = seqnums[fd]++;

     if (h->opcode >= nfuncs || h->opcode < 0) {
	  ARLADEB (ADEBMISC, ("Bad message opcode = %u\n", h->opcode));
	  return -1;
     }

     ARLADEB (ADEBMISC, ("Send message: opcode = %u (%s), size = %u\n",
	      h->opcode, rcvfuncs_name[h->opcode], h->size));

     if ((res = write (fd, h, size)) < 0) {
	  perror ("write");
	  ARLADEB (ADEBMISC, ("xfs_message_send: write: fd: %d, size: %d\n", 
		   fd, size));
	  return -1;
     } else
	  return 0;
}

int
xfs_message_wakeup (int fd, struct xfs_message_wakeup *h, u_int size)
{
     Listitem *i;
     struct xfs_message_header *w;

     for (i = listhead (sleepers); i; i = listnext (sleepers, i)) {
	  w = (struct xfs_message_header *)listdata(i);
	  if (w->sequence_num == h->sleepers_sequence_num) {
	       listdel (sleepers, i);
	       memcpy (w, h, size);
	       LWP_SignalProcess ((char *)w);
	  }
     }
     return 0;
}

static int
xfs_message_sleep (struct xfs_message_header *h)
{
     listaddtail (sleepers, h);
     LWP_WaitProcess ((char *)h);
     return ((struct xfs_message_wakeup *)h)->errno;
}

int
xfs_message_rpc (int fd, struct xfs_message_header *h, u_int size)
{
     if (size < sizeof (struct xfs_message_wakeup)) {
	  ARLADEB (ADEBMISC, ("xfs_message_rpc: Too small packet for rpc\n"));
	  return -1;
     }
     return xfs_message_send (fd, h, size) || xfs_message_sleep (h);
}

static int
xfs_send_message_wakeup (int fd, u_int seqnum, int errno)
{
     struct xfs_message_wakeup msg;
     
     msg.header.opcode = XFS_MSG_WAKEUP;
     msg.sleepers_sequence_num = seqnum;
     msg.errno = errno;
     ARLADEB (ADEBMISC, ("sending wakeup: seq = %u, errno = %d\n", seqnum, errno));
     return xfs_message_send (fd, (struct xfs_message_header *)&msg, 
			      sizeof(msg));
}

static int
xfs_send_message_wakeup_data (int fd, u_int seqnum, int errno,
			      void *data, int size)
{
     struct xfs_message_wakeup_data msg;
     
     msg.header.opcode = XFS_MSG_WAKEUP_DATA;
     msg.sleepers_sequence_num = seqnum;
     msg.errno = errno;
     ARLADEB (ADEBMISC, ("sending wakeup: seq = %u, errno = %d\n", seqnum, errno));

     if (sizeof(msg) >= size && size != 0) {
	 memcpy(msg.msg, data, size);
     }

     msg.len = size;

     return xfs_message_send (fd, (struct xfs_message_header *)&msg, 
			      sizeof(msg));
}

int
xfs_message_getroot (int fd, struct xfs_message_getroot *h, u_int size)
{
     struct xfs_message_installroot msg;
     int ret = 0;
     VenusFid root_fid;
     VenusFid real_fid;
     AFSFetchStatus status;
     Result result;
     CredCacheEntry *ce;

     ce = cred_get (0, h->cred.pag, CRED_ANY);
     assert (ce != NULL);

     ret = getroot (&root_fid, ce);
     if (ret)
	 goto out;

     result = cm_getattr(root_fid, &status, &real_fid, ce);
     if (result.res == -1) {
	 ret = result.errno;
	 goto out;
     }	 

     fcacheentry2xfsnode (&root_fid, &real_fid,
			  &status, &msg.node);

     msg.header.opcode = XFS_MSG_INSTALLROOT;
     xfs_message_send (fd, (struct xfs_message_header *)&msg, sizeof(msg));

out:
     cred_free (ce);

     xfs_send_message_wakeup (fd, h->header.sequence_num, ret);
     return ret;
}

int
xfs_message_getnode (int fd, struct xfs_message_getnode *h, u_int size)
{
     struct xfs_message_installnode msg;
     VenusFid *dirfid = (VenusFid *)&h->parent_handle;
     VenusFid fid;
     VenusFid real_fid;
     Result res;
     AFSFetchStatus status;
     CredCacheEntry *ce;

     ce = cred_get (dirfid->Cell, h->cred.pag, CRED_ANY);
     assert (ce != NULL);

     res = cm_lookup (*dirfid, h->name, &fid, &ce);
     if (res.res == 0) {
	  res = cm_getattr (fid, &status, &real_fid, ce);
	  if (res.res == 0) {
	       fcacheentry2xfsnode (&fid, &real_fid,
				    &status, &msg.node);

	       msg.node.tokens = res.tokens & ~XFS_DATA_MASK;
	       msg.parent_handle = h->parent_handle;
	       strcpy (msg.name, h->name);

	       msg.header.opcode = XFS_MSG_INSTALLNODE;
	       xfs_message_send (fd, (struct xfs_message_header *)&msg, 
				 sizeof(msg));
	  }
     }
     cred_free (ce);
     xfs_send_message_wakeup (fd, h->header.sequence_num, 
			      res.res == -1 ? res.errno : res.res);
     return 0;
}

int
xfs_message_getattr (int fd, struct xfs_message_getattr *h, u_int size)
{
     struct xfs_message_installattr msg;
     VenusFid *fid;
     VenusFid real_fid;
     AFSFetchStatus status;
     Result res;
     CredCacheEntry *ce;

     fid = (VenusFid *)&h->handle;

     ce = cred_get (fid->Cell, h->cred.pag, CRED_ANY);
     assert (ce != NULL);

     res = cm_getattr (*fid, &status, &real_fid, ce);
     if (res.res == 0) {
	 fcacheentry2xfsnode (fid, &real_fid,
			      &status, &msg.node);

	  msg.node.tokens = res.tokens & ~XFS_DATA_MASK;
	  msg.header.opcode = XFS_MSG_INSTALLATTR;
	  xfs_message_send (fd, (struct xfs_message_header *)&msg, 
			    sizeof(msg));
     }

     cred_free (ce);
     xfs_send_message_wakeup (fd, h->header.sequence_num, 
			      res.res ? res.errno : res.res);
     return 0;
}


int 
xfs_message_putattr (int fd, struct xfs_message_putattr *h, u_int size)
{
     VenusFid *fid;
     AFSStoreStatus status;
     Result res;
     CredCacheEntry *ce;

     fid = (VenusFid *)&h->handle;

     ce = cred_get (fid->Cell, h->cred.pag, CRED_ANY);
     assert (ce != NULL);

     xfs_attr2afsstorestatus(&h->attr, &status);
     res.res = 0;
     if (XA_VALID_SIZE(&h->attr))
	 res = cm_ftruncate (*fid, h->attr.xa_size, ce);

     if (res.res == 0)
	 res = cm_setattr(*fid, &status, ce);

     cred_free (ce);
     xfs_send_message_wakeup (fd, h->header.sequence_num, 
			      res.res ? res.errno : res.res);
     return 0;
}

int 
xfs_message_create (int fd, struct xfs_message_create *h, u_int size)
{
     VenusFid *parent_fid, child_fid;
     AFSStoreStatus store_status;
     AFSFetchStatus fetch_status;
     Result res;
     CredCacheEntry *ce;
     int ret;

     parent_fid = (VenusFid *)&h->parent_handle;

     ce = cred_get (parent_fid->Cell, h->cred.pag, CRED_ANY);
     assert (ce != NULL);

     xfs_attr2afsstorestatus(&h->attr, &store_status);
     res = cm_create(*parent_fid, h->name, &store_status,
		     &child_fid, &fetch_status, ce);
     if (res.res == 0) {
	 struct xfs_message_installdata msg1;
	 struct xfs_message_installnode msg2;
	 FCacheEntry *dir_entry;
	 char tmp[5];

	 ret = fcache_get (&dir_entry, *parent_fid, ce);
	 if (ret)
	     goto out;

	 ret = fcache_get_data (dir_entry, ce);
	 if (ret) {
	     ReleaseWriteLock (&dir_entry->lock);
	     goto out;
	 }
	     
	 res = conv_dir (dir_entry, tmp, ce, 0); /* XXX */
	 if (res.res == -1) {
	     ReleaseWriteLock (&dir_entry->lock);
	     ret = res.errno;
	     goto out;
	 }
	 strncpy((char *)&msg1.cache_handle, tmp, 4);
	 msg1.node.tokens = res.tokens;

	 fcacheentry2xfsnode (parent_fid, &dir_entry->realfid,
			      &dir_entry->status, &msg1.node);
	 
	 msg1.header.opcode = XFS_MSG_INSTALLDATA;
	 xfs_message_send (fd, (struct xfs_message_header *)&msg1,
			   sizeof(msg1));
	 ReleaseWriteLock (&dir_entry->lock);

	 fcacheentry2xfsnode (&child_fid, &child_fid,
			      &fetch_status, &msg2.node);
			      
	 msg2.node.tokens   = XFS_ATTR_R; /* XXX */
	 msg2.parent_handle = h->parent_handle;
	 strcpy (msg2.name, h->name);

	 msg2.header.opcode = XFS_MSG_INSTALLNODE;
	 xfs_message_send (fd, (struct xfs_message_header *)&msg2,
			   sizeof(msg2));
     }

     if (res.res == -1)
	 ret = res.errno;
     else
	 ret = 0;
out:
     cred_free (ce);
     xfs_send_message_wakeup (fd, h->header.sequence_num, ret);
     return ret;
}

int 
xfs_message_mkdir (int fd, struct xfs_message_mkdir *h, u_int size)
{
     VenusFid *parent_fid, child_fid;
     AFSStoreStatus store_status;
     AFSFetchStatus fetch_status;
     Result res;
     CredCacheEntry *ce;
     int ret;

     parent_fid = (VenusFid *)&h->parent_handle;

     ce = cred_get (parent_fid->Cell, h->cred.pag, CRED_ANY);
     assert (ce != NULL);

     xfs_attr2afsstorestatus(&h->attr, &store_status);
     res = cm_mkdir(*parent_fid, h->name, &store_status,
		    &child_fid, &fetch_status, ce);
     if (res.res == 0) {
	 struct xfs_message_installdata msg1;
	 struct xfs_message_installnode msg2;
	 FCacheEntry *dir_entry;
	 char tmp[5];

	 ret = fcache_get (&dir_entry, *parent_fid, ce);
	 if (ret)
	     goto out;

	 ret = fcache_get_data (dir_entry, ce);
	 if (ret) {
	     ReleaseWriteLock (&dir_entry->lock);
	     goto out;
	 }
	     
	 res = conv_dir (dir_entry, tmp, ce, 0); /* XXX */
	 if (res.res == -1) {
	     ReleaseWriteLock (&dir_entry->lock);
	     ret = res.errno;
	     goto out;
	 }
	 strncpy((char *)&msg1.cache_handle, tmp, 4);
	 msg1.node.tokens = res.tokens;

	 fcacheentry2xfsnode (parent_fid, &dir_entry->realfid,
			      &dir_entry->status, &msg1.node);
	 
	 msg1.header.opcode = XFS_MSG_INSTALLDATA;
	 xfs_message_send (fd, (struct xfs_message_header *)&msg1,
			   sizeof(msg1));
	 ReleaseWriteLock (&dir_entry->lock);

	 fcacheentry2xfsnode (&child_fid, &child_fid,
			      &fetch_status, &msg2.node);
			      
	 msg2.node.tokens   = XFS_ATTR_R; /* XXX */
	 msg2.parent_handle = h->parent_handle;
	 strcpy (msg2.name, h->name);

	 msg2.header.opcode = XFS_MSG_INSTALLNODE;
	 xfs_message_send (fd, (struct xfs_message_header *)&msg2,
			   sizeof(msg2));
     }

     if (res.res == -1)
	 ret = res.errno;
     else
	 ret = 0;
out:
     cred_free (ce);
     xfs_send_message_wakeup (fd, h->header.sequence_num, ret);
     return ret;
}

int 
xfs_message_link (int fd, struct xfs_message_link *h, u_int size)
{
     VenusFid *parent_fid, *existing_fid;
     AFSFetchStatus fetch_status;
     Result res;
     CredCacheEntry *ce;
     int ret;

     parent_fid   = (VenusFid *)&h->parent_handle;
     existing_fid = (VenusFid *)&h->from_handle;

     ce = cred_get (parent_fid->Cell, h->cred.pag, CRED_ANY);
     assert (ce != NULL);

     res = cm_link (*parent_fid, h->name, *existing_fid,
		    &fetch_status, ce);
     if (res.res == 0) {
	 struct xfs_message_installdata msg1;
	 struct xfs_message_installnode msg2;
	 FCacheEntry *dir_entry;
	 char tmp[5];

	 ret = fcache_get (&dir_entry, *parent_fid, ce);
	 if (ret)
	     goto out;

	 ret = fcache_get_data (dir_entry, ce);
	 if (ret) {
	     ReleaseWriteLock (&dir_entry->lock);
	     goto out;
	 }
	     
	 res = conv_dir (dir_entry, tmp, ce, 0); /* XXX */
	 if (res.res == -1) {
	     ReleaseWriteLock (&dir_entry->lock);
	     ret = res.errno;
	     goto out;
	 }
	 strncpy((char *)&msg1.cache_handle, tmp, 4);
	 msg1.node.tokens = res.tokens;

	 fcacheentry2xfsnode (parent_fid, &dir_entry->realfid,
			      &dir_entry->status, &msg1.node);
	 
	 msg1.header.opcode = XFS_MSG_INSTALLDATA;
	 xfs_message_send (fd, (struct xfs_message_header *)&msg1,
			   sizeof(msg1));
	 ReleaseWriteLock (&dir_entry->lock);

	 fcacheentry2xfsnode (existing_fid, existing_fid,
			      &fetch_status, &msg2.node);
			      
	 msg2.node.tokens   = XFS_ATTR_R; /* XXX */
	 msg2.parent_handle = h->parent_handle;
	 strcpy (msg2.name, h->name);

	 msg2.header.opcode = XFS_MSG_INSTALLNODE;
	 xfs_message_send (fd, (struct xfs_message_header *)&msg2,
			   sizeof(msg2));
     }

     if (res.res == -1)
	 ret = res.errno;
     else
	 ret = 0;
out:
     cred_free (ce);
     xfs_send_message_wakeup (fd, h->header.sequence_num, ret);
     return ret;
}

int 
xfs_message_symlink (int fd, struct xfs_message_symlink *h, u_int size)
{
     VenusFid *parent_fid, child_fid;
     AFSStoreStatus store_status;
     AFSFetchStatus fetch_status;
     Result res;
     CredCacheEntry *ce;
     int ret;

     parent_fid = (VenusFid *)&h->parent_handle;

     ce = cred_get (parent_fid->Cell, h->cred.pag, CRED_ANY);
     assert (ce != NULL);

     xfs_attr2afsstorestatus(&h->attr, &store_status);
     res = cm_symlink(*parent_fid, h->name, &store_status,
		      &child_fid, &fetch_status,
		      h->contents, ce);
     if (res.res == 0) {
	 struct xfs_message_installdata msg1;
	 struct xfs_message_installnode msg2;
	 FCacheEntry *dir_entry;
	 char tmp[5];

	 ret = fcache_get (&dir_entry, *parent_fid, ce);
	 if (ret)
	     goto out;

	 ret = fcache_get_data (dir_entry, ce);
	 if (ret) {
	     ReleaseWriteLock (&dir_entry->lock);
	     goto out;
	 }
	     
	 res = conv_dir (dir_entry, tmp, ce, 0); /* XXX */
	 if (res.res == -1) {
	     ReleaseWriteLock (&dir_entry->lock);
	     ret = res.errno;
	     goto out;
	 }
	 strncpy((char *)&msg1.cache_handle, tmp, 4);
	 msg1.node.tokens = res.tokens;

	 fcacheentry2xfsnode (parent_fid, &dir_entry->realfid,
			      &dir_entry->status, &msg1.node);
	 
	 msg1.header.opcode = XFS_MSG_INSTALLDATA;
	 xfs_message_send (fd, (struct xfs_message_header *)&msg1,
			   sizeof(msg1));
	 ReleaseWriteLock (&dir_entry->lock);

	 fcacheentry2xfsnode (&child_fid, &child_fid,
			      &fetch_status, &msg2.node);
			      
	 msg2.node.tokens   = XFS_ATTR_R; /* XXX */
	 msg2.parent_handle = h->parent_handle;
	 strcpy (msg2.name, h->name);

	 msg2.header.opcode = XFS_MSG_INSTALLNODE;
	 xfs_message_send (fd, (struct xfs_message_header *)&msg2,
			   sizeof(msg2));
     }

     if (res.res == -1)
	 ret = res.errno;
     else
	 ret = 0;
out:
     cred_free (ce);
     xfs_send_message_wakeup (fd, h->header.sequence_num, ret);
     return ret;
}

int 
xfs_message_remove (int fd, struct xfs_message_remove *h, u_int size)
{
     VenusFid *parent_fid;
     Result res;
     CredCacheEntry *ce;
     int ret;

     parent_fid = (VenusFid *)&h->parent_handle;

     ce = cred_get (parent_fid->Cell, h->cred.pag, CRED_ANY);
     assert (ce != NULL);

     res = cm_remove(*parent_fid, h->name, ce);
     if (res.res == 0) {
	 struct xfs_message_installdata msg;
	 FCacheEntry *dir_entry;
	 char tmp[5];

	 ret = fcache_get (&dir_entry, *parent_fid, ce);
	 if (ret)
	     goto out;

	 ret = fcache_get_data (dir_entry, ce);
	 if (ret) {
	     ReleaseWriteLock (&dir_entry->lock);
	     goto out;
	 }
	     
	 res = conv_dir (dir_entry, tmp, ce, 0); /* XXX */
	 if (res.res == -1) {
	     ReleaseWriteLock (&dir_entry->lock);
	     ret = res.errno;
	     goto out;
	 }
	 strncpy((char *)&msg.cache_handle, tmp, 4);
	 msg.node.tokens = res.tokens;

	 fcacheentry2xfsnode (parent_fid, &dir_entry->realfid,
			      &dir_entry->status, &msg.node);
	 
	 msg.header.opcode = XFS_MSG_INSTALLDATA;
	 xfs_message_send (fd, (struct xfs_message_header *)&msg,
			   sizeof(msg));
	 ReleaseWriteLock (&dir_entry->lock);
     }

     if (res.res == -1)
	 ret = res.errno;
     else
	 ret = 0;
out:
     cred_free (ce);
     xfs_send_message_wakeup (fd, h->header.sequence_num, ret);
     return ret;
}

int 
xfs_message_rmdir (int fd, struct xfs_message_rmdir *h, u_int size)
{
     VenusFid *parent_fid;
     Result res;
     CredCacheEntry *ce;
     int ret;

     parent_fid = (VenusFid *)&h->parent_handle;

     ce = cred_get (parent_fid->Cell, h->cred.pag, CRED_ANY);
     assert (ce != NULL);

     res = cm_rmdir(*parent_fid, h->name, ce);
     if (res.res == 0) {
	 struct xfs_message_installdata msg;
	 FCacheEntry *dir_entry;
	 char tmp[5];

	 ret = fcache_get (&dir_entry, *parent_fid, ce);
	 if (ret)
	     goto out;

	 ret = fcache_get_data (dir_entry, ce);
	 if (ret) {
	     ReleaseWriteLock (&dir_entry->lock);
	     goto out;
	 }
	     
	 res = conv_dir (dir_entry, tmp, ce, 0); /* XXX */
	 if (res.res == -1) {
	     ReleaseWriteLock (&dir_entry->lock);
	     ret = res.errno;
	     goto out;
	 }
	 strncpy((char *)&msg.cache_handle, tmp, 4);
	 msg.node.tokens = res.tokens;

	 fcacheentry2xfsnode (parent_fid, &dir_entry->realfid,
			      &dir_entry->status, &msg.node);
	 
	 msg.header.opcode = XFS_MSG_INSTALLDATA;
	 xfs_message_send (fd, (struct xfs_message_header *)&msg,
			   sizeof(msg));
	 ReleaseWriteLock (&dir_entry->lock);
     }

     if (res.res == -1)
	 ret = res.errno;
     else
	 ret = 0;
out:
     cred_free (ce);
     xfs_send_message_wakeup (fd, h->header.sequence_num, ret);
     return ret;
}

int 
xfs_message_rename (int fd, struct xfs_message_rename *h, u_int size)
{
     VenusFid *old_parent_fid;
     VenusFid *new_parent_fid;
     Result res;
     CredCacheEntry *ce;
     int ret;

     old_parent_fid = (VenusFid *)&h->old_parent_handle;
     new_parent_fid = (VenusFid *)&h->new_parent_handle;

     ce = cred_get (old_parent_fid->Cell, h->cred.pag, CRED_ANY);
     assert (ce != NULL);

     res = cm_rename(*old_parent_fid, h->old_name,
		     *new_parent_fid, h->new_name,
		     ce);

     if (res.res == 0) {
	 struct xfs_message_installdata msg;
	 FCacheEntry *dir_entry;
	 char tmp[5];

	 ret = fcache_get (&dir_entry, *old_parent_fid, ce);
	 if (ret)
	     goto out;

	 ret = fcache_get_data (dir_entry, ce);
	 if (ret) {
	     ReleaseWriteLock (&dir_entry->lock);
	     goto out;
	 }
	     
	 res = conv_dir (dir_entry, tmp, ce, 0); /* XXX */
	 if (res.res == -1) {
	     ReleaseWriteLock (&dir_entry->lock);
	     ret = res.errno;
	     goto out;
	 }
	 strncpy((char *)&msg.cache_handle, tmp, 4);
	 msg.node.tokens = res.tokens;

	 fcacheentry2xfsnode (old_parent_fid, &dir_entry->realfid,
			      &dir_entry->status, &msg.node);
	 
	 msg.header.opcode = XFS_MSG_INSTALLDATA;
	 xfs_message_send (fd, (struct xfs_message_header *)&msg,
			   sizeof(msg));
	 ReleaseWriteLock (&dir_entry->lock);

	 /* new parent */

	 ret = fcache_get (&dir_entry, *new_parent_fid, ce);
	 if (ret)
	     goto out;

	 ret = fcache_get_data (dir_entry, ce);
	 if (ret) {
	     ReleaseWriteLock (&dir_entry->lock);
	     goto out;
	 }
	     
	 res = conv_dir (dir_entry, tmp, ce, 0); /* XXX */
	 if (res.res == -1) {
	     ReleaseWriteLock (&dir_entry->lock);
	     ret = res.errno;
	     goto out;
	 }
	 strncpy((char *)&msg.cache_handle, tmp, 4);
	 msg.node.tokens = res.tokens;

	 fcacheentry2xfsnode (new_parent_fid, &dir_entry->realfid,
			      &dir_entry->status, &msg.node);
	 
	 msg.header.opcode = XFS_MSG_INSTALLDATA;
	 xfs_message_send (fd, (struct xfs_message_header *)&msg,
			   sizeof(msg));
	 ReleaseWriteLock (&dir_entry->lock);
     }

     if (res.res == -1)
	 ret = res.errno;
     else
	 ret = 0;
out:
     cred_free (ce);
     xfs_send_message_wakeup (fd, h->header.sequence_num, ret);
     return ret;
}

int 
xfs_message_putdata (int fd, struct xfs_message_putdata *h, u_int size)
{
     AFSFetchStatus status;
     VenusFid *fid;
     VenusFid real_fid;
     Result res;
     CredCacheEntry *ce;

     fid = (VenusFid *)&h->handle;

     ce = cred_get (fid->Cell, h->cred.pag, CRED_ANY);
     assert (ce != NULL);

     res = cm_getattr (*fid, &status, &real_fid, ce);
     if (res.res == 0) {
	 res = cm_close(*fid, h->flag, ce);
	 if (res.res != 0)
	     ARLADEB (ADEBMISC, ("xfs_message_putdata: cm_close failed %d\n",
		     res.errno));
     }

     cred_free (ce);
     xfs_send_message_wakeup (fd, h->header.sequence_num,
			      res.res ? res.errno : res.res);
     return 0;
}

int
xfs_message_getdata (int fd, struct xfs_message_getdata *h, u_int size)
{
     struct xfs_message_installdata msg;
     VenusFid *fid;
     VenusFid real_fid;
     Result res;
     AFSFetchStatus status;
     char tmp[5];
     CredCacheEntry *ce;
     int ret;

     fid = (VenusFid *)&h->handle;

     ce = cred_get (fid->Cell, h->cred.pag, CRED_ANY);
     assert (ce != NULL);

     ARLADEB (ADEBMISC, ("open: before cm_getattr\n"));
     res = cm_getattr (*fid, &status, &real_fid, ce);
     ARLADEB (ADEBMISC, ("open: after cm_getattr\n"));
     if (res.res == 0) {
	  fcacheentry2xfsnode (fid, &real_fid, &status, &msg.node);
	  if (status.FileType == TYPE_DIR) {
	       FCacheEntry *entry;

	       ARLADEB (ADEBMISC, ("open: before fcache_get\n"));
	       ret = fcache_get (&entry, *fid, ce);
	       if (ret)
		   goto out;

	       ARLADEB (ADEBMISC, ("open: before fcache_get_data\n"));
	       ret = fcache_get_data (entry, ce);
	       if (ret) {
		   ReleaseWriteLock (&entry->lock);
		   goto out;
	       }

	       ARLADEB (ADEBMISC, ("open: before conv_dir\n"));
	       res = conv_dir (entry, tmp, ce, h->tokens);
	       if (res.res != -1) {
		    strncpy ((char *)&msg.cache_handle, tmp, 4);
		    msg.node.tokens = res.tokens;
	       }
	       ReleaseWriteLock(&entry->lock);
	  } else {
	       ARLADEB (ADEBMISC, ("open: before cm_open\n"));
	       res = cm_open (*fid, ce, h->tokens);
	       ARLADEB (ADEBMISC, ("open: after cm_open\n"));
	       if (res.res != -1) {
		    sprintf (tmp, "%04X", res.res);
		    strncpy ((char *)&msg.cache_handle, tmp, 4);
		    msg.node.tokens = res.tokens;
	       }
	  }
     }

     if (res.res != -1) {
	  msg.header.opcode = XFS_MSG_INSTALLDATA;
	  xfs_message_send (fd, (struct xfs_message_header *)&msg, 
			    sizeof(msg));
     }
     if (res.res == -1)
	 ret = res.errno;
     else
	 ret = 0;

out:
     cred_free (ce);
     xfs_send_message_wakeup (fd, h->header.sequence_num, ret);
     return ret;
}

void
break_callback (VenusFid fid)
{
     struct xfs_message_invalidnode msg;

     msg.header.opcode = XFS_MSG_INVALIDNODE;
     memcpy (&msg.handle, &fid, sizeof(fid));
     xfs_message_send (kernel_fd , (struct xfs_message_header *)&msg, 
		       sizeof(msg));
}

int
xfs_message_inactivenode (int fd, struct xfs_message_inactivenode *h, 
			  u_int size)
{
     FCacheEntry *entry;
     VenusFid *fid;
     int ret;
     CredCacheEntry *ce;

     fid = (VenusFid *)&h->handle;

     ce = cred_get (fid->Cell, 0, CRED_NONE);
     assert (ce != NULL);

     ret = fcache_get (&entry, *fid, ce);
     cred_free (ce);

#if 0
     ret = fcache_get_attr (entry, ce);
#endif
     if (ret) {
	 ARLADEB (ADEBMISC, ("xfs_message_inactivenode: node not found\n"));
	 return 0;
     }
     entry->flags.datausedp = entry->flags.attrusedp = FALSE;
     ReleaseWriteLock (&entry->lock);
     return 0;
}

/*
 * Handle VIOCGETTOK
 */

#ifdef KERBEROS
static int
token_for_cell (int fd, struct xfs_message_pioctl *h, u_int size,
		CredCacheEntry *ce)
{
    struct ClearToken ct;
    char buf[2048];
    size_t len;
    char *p = buf;
    u_int32_t tmp;
    krbstruct *kstruct = (krbstruct *)ce->cred_data;
    CREDENTIALS *cred  = &kstruct->c;
    const char *cell = cell_num2name (ce->cell);

    ct.AuthHandle = cred->kvno;
    memcpy (ct.HandShakeKey, cred->session, sizeof(cred->session));
    ct.ViceId = h->cred.pag;
    ct.BeginTimestamp = cred->issue_date + 1;
    ct.EndTimestamp   = ce->expire;

    tmp = 0;
    memcpy (p, &tmp, sizeof(tmp));
    p += sizeof(tmp);
    tmp = sizeof(ct);
    memcpy (p, &tmp, sizeof(tmp));
    p += sizeof(tmp);
    memcpy (p, &ct, sizeof(ct));
    p += sizeof(ct);
    tmp = strlen(cell);
    memcpy (p, &tmp, sizeof(tmp));
    p += sizeof(tmp);
    strcpy (p, cell);
    p += strlen(cell) + 1;

    len = p - buf;

    memset (&ct, 0, sizeof(ct));

    cred_free (ce);

    xfs_send_message_wakeup_data (fd, h->header.sequence_num, 0,
				  buf, len);
    return 0;
}

static int
viocflush (int fd, struct xfs_message_pioctl *h, u_int size)
{
    VenusFid fid ;
    AFSCallBack broken_callback = {0, 0, CBDROPPED};

    fid.Cell = h->handle.a;
    fid.fid.Volume = h->handle.b;
    fid.fid.Vnode = h->handle.c;
    fid.fid.Unique = h->handle.d;

    ARLADEB(ADEBMISC,
	    ("flushing (%d, %u, %u, %u)\n",
	     fid.Cell, fid.fid.Volume, fid.fid.Vnode, fid.fid.Unique));

    fcache_stale_entry(fid, broken_callback);
    return 0 ;
}


static int
viocgettok (int fd, struct xfs_message_pioctl *h, u_int size)
{
    if (h->insize == 0) {
	CredCacheEntry *ce = cred_get (0, h->cred.pag, CRED_KRB4);

	if (ce == NULL) {
	    xfs_send_message_wakeup (fd, h->header.sequence_num, ENOTCONN);
	    return 0;
	}
	return token_for_cell (fd, h, size, ce);
    } else if (h->insize == sizeof(u_int32_t)) {
	u_int32_t n;
	int i, c;
	int found;
	CredCacheEntry *ce;

	memcpy (&n, h->msg, sizeof(n));

	i = 0;
	c = 0;
	found = 0;
	while (!found && i <= n) {
	    if (cell_num2name(c) == NULL)
		break;

	    ce = cred_get (c++, h->cred.pag, CRED_KRB4);
	    if (ce != NULL) {
		if (i == n)
		    found = 1;
		else
		    ++i;
	    }
	}
	if (!found) {
	    xfs_send_message_wakeup (fd, h->header.sequence_num, EDOM);
	    return 0;
	}
	return token_for_cell (fd, h, size, ce);
    } else {
	xfs_send_message_wakeup (fd, h->header.sequence_num, EINVAL);
    }
    return 0;
}
#endif /* KERBEROS */

int
xfs_message_pioctl (int fd, struct xfs_message_pioctl *h, u_int size)
{
    int32_t sizeof_x;
    char *t;

    t = h->msg ;
    switch(h->opcode) {
#ifdef KERBEROS
    case VIOCSETTOK: {
	struct ClearToken ct;
	CREDENTIALS c;
	long cell;

	/* someone probed us */
	if (h->insize == 0) {
	    errno = EINVAL ;
	    break;
	}

	/* Get ticket_st */
	memcpy(&sizeof_x, t, sizeof(sizeof_x)) ;
	c.ticket_st.length = sizeof_x ;
	ARLADEB (ADEBMISC, ("ticket_st has size %d\n", sizeof_x));
	t += sizeof(sizeof_x) ;

	memcpy(c.ticket_st.dat, t, sizeof_x) ;
	t += sizeof_x ;

	/* Get ClearToken */
	memcpy(&sizeof_x, t, sizeof(sizeof_x)) ;
	t += sizeof(sizeof_x) ;

	memcpy(&ct, t, sizeof_x) ;
	t += sizeof_x ;

	/* Get primary cell ? */
	memcpy(&sizeof_x, t, sizeof(sizeof_x)) ;
	t += sizeof(sizeof_x) ;

	/* Get Cellname */ 
	strncpy(c.realm, t, REALM_SZ) ;
	c.realm[REALM_SZ-1] = '\0' ;


	/* Make this a sane world again */
	c.kvno = ct.AuthHandle;
	memcpy (c.session, ct.HandShakeKey, sizeof(c.session));
	c.issue_date = ct.BeginTimestamp - 1;
	
	cell = cell_name2num(strlwr(c.realm));

	/* XXX fix ct.ViceId */
	cred_add (ct.ViceId, CRED_KRB4, 2, cell, ct.EndTimestamp,
		  &c, sizeof(c));
	
	conn_clearcred (cell, h->cred.pag, 2);

	errno = 0 ;
	break;
    }
    case VIOCGETTOK :
	return viocgettok (fd, h, size);
    case VIOCFLUSH:
	return viocflush(fd, h, size);
    case VIOC_VENUSLOG:
	if (h->cred.uid != 0) {
	    errno = EACCES;
	    break ;
	}
	    
	conn_status (stderr);
	volcache_status (stderr);
	cred_status (stderr);
	fcache_status (stderr);
	rx_PrintStats(stderr);
	errno = 0 ;
	break;
    case VIOCUNLOG: {
	pag_t cred = h->cred.pag ;

	cred_remove(cred) ;
	errno = 0;
	break ;
    }
#endif /* KERBEROS */
    case VIOC_AFS_SYSNAME: {
	char str[SYSNAMEMAXLEN+sizeof(int32_t)];

	errno = 0 ;

	if (*((int32_t *)t)) {
	    t += sizeof(int32_t);
	    ARLADEB (ADEBMISC, ("VIOC_AFS_SYSNAME: setting sysname: %s\n", t));
	    memcpy(arlasysname, t, h->insize);
	    arlasysname[h->insize] = '\0';
	    xfs_send_message_wakeup_data(fd, h->header.sequence_num, errno,
					 str, 0);
	} else {
	    t = str;
	    sizeof_x = strlen(arlasysname);
	    memcpy(t, &sizeof_x, sizeof(sizeof_x));
	    t += sizeof(sizeof_x);
	    h->outsize = sizeof_x;
	    strncpy(t, arlasysname, SYSNAMEMAXLEN);
	    xfs_send_message_wakeup_data(fd, h->header.sequence_num, errno,
					 str, sizeof_x + sizeof(sizeof_x));
	}
	return 0;
    }

    case VIOC_FILE_CELL_NAME: {
	char *cellname ;

	errno = 0 ;
	cellname = (char *) cell_num2name(h->handle.a);

	if (cellname) 
	    xfs_send_message_wakeup_data(fd, h->header.sequence_num, errno,
					 cellname, strlen(cellname));
	else 
	    xfs_send_message_wakeup_data(fd, h->header.sequence_num, EINVAL,
					 NULL, 0);
	return 0;
    }
    case VIOC_GET_WS_CELL: {
	char *cellname;

	cellname = (char*) cell_getthiscell();
	xfs_send_message_wakeup_data(fd, h->header.sequence_num, 0 /*errno*/,
				     cellname, strlen(cellname));

	return 0;
    }
    case VIOCSETCACHESIZE: {
	u_int32_t *s = (u_int32_t *)t;
	
	if (h->insize >= sizeof(int32_t) * 4) 
	    errno = fcache_reinit(s[0], s[1], s[2], s[3]);
	else
	    errno = fcache_reinit(*s/2, *s, *s*500, *s*1000);
	break;
    }
    default:
	ARLADEB (ADEBMISC, ("unknown pioctl call\n"));
	errno = EINVAL ;
    }

    xfs_send_message_wakeup (fd, h->header.sequence_num, errno);
    
    return 0;
}
