/*
 * Copyright (c) 1995, 1996, 1997, 1998 Kungliga Tekniska Hgskolan
 * (Royal Institute of Technology, Stockholm, Sweden).
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Kungliga Tekniska
 *      Hgskolan and its contributors.
 *
 * 4. Neither the name of the Institute nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "xfs_locl.h"
#include <xfs_deb.h>
#include <xfs_fs.h>
#include <xfs_message.h>	      

#ifdef 0			       /* XXX - remove */
#include <sys/ucred.h>
#endif				       /* 0 */

RCSID("$Id: xfs_message.c,v 1.7 1998/02/01 21:44:26 art Exp $");

int
xfs_message_installroot(int fd,
			struct xfs_message_installroot * message,
			u_int size,
			struct proc * p)
{
    int error = 0;

    XFSDEB(XDEBMSG, ("xfs_message_installroot\n"));

    if (xfs[fd].root != NULL) {
	printf("XFS PANIC Warning: xfs_message_installroot again\n");
	error = EBUSY;
    } else {
	error = new_xfs_node(&xfs[fd], &message->node, &xfs[fd].root, p);
	if (error)
	    return error;
	xfs[fd].root->vn->v_flag |= VROOT;

    }
    return error;
}

int
xfs_message_installnode(int fd,
			struct xfs_message_installnode * message,
			u_int size,
			struct proc * p)
{
    int error = 0;
    struct xfs_node *n, *dp;

    XFSDEB(XDEBMSG, ("xfs_message_installnode\n"));

    dp = xfs_node_find(&xfs[fd], &message->parent_handle);
    if (dp) {
	error = new_xfs_node(&xfs[fd], &message->node, &n, p);
	if (error)
	    return error;
	xfs_dnlc_enter(XNODE_TO_VNODE(dp), message->name, XNODE_TO_VNODE(n));
	vrele(XNODE_TO_VNODE(n));
    } else {
	printf("XFS PANIC Warning: xfs_message_install could not "
	       "find parent\n");
	error = ENOENT;
    }

    return error;
}

int
xfs_message_installattr(int fd,
			struct xfs_message_installattr * message,
			u_int size)
{
    int errno = 0;
    struct xfs_node *t;

    t = xfs_node_find(&xfs[fd], &message->node.handle);
    if (t != 0) {
	t->tokens = message->node.tokens;
	xfs_attr2vattr(&message->node.attr, &t->attr);
	bcopy(message->node.id, t->id, sizeof(t->id));
	bcopy(message->node.rights, t->rights, sizeof(t->rights));
    } else {
	printf("XFS PANIC Warning: xfs_message_installattr didn't find node!\n");
	errno = ENOENT;
    }

    return errno;
}

int
xfs_message_installdata(int fd,
			struct xfs_message_installdata * message,
			u_int size,
			struct proc * p)
{
    struct xfs_node *t;
    int errno = 0;

    XFSDEB(XDEBMSG, ("xfs_message_installdata\n"));

    t = xfs_node_find(&xfs[fd], &message->node.handle);
    if (t != NULL) {
	struct componentname cn;
	char tmp[sizeof(message->cache_handle) + 1];
	char *cp;
	struct vnode *cwd;
	struct vnode *vp;

	XFSDEB(XDEBMSG, ("cache_handle = '%c%c%c%c'\n",
			 (int) message->cache_handle.data[0],
			 (int) message->cache_handle.data[1],
			 (int) message->cache_handle.data[2],
			 (int) message->cache_handle.data[3]));

	bcopy((char *) &message->cache_handle,
	      tmp,
	      sizeof(message->cache_handle));
	tmp[4] = '\0';

	cn.cn_nameiop = LOOKUP;
	cn.cn_flags = LOCKLEAF;
	cn.cn_proc = curproc;
	cn.cn_namelen = strlen(tmp);
	cn.cn_nameptr = tmp;
	cn.cn_hash = 0;
	for (cp = tmp; *cp; ++cp)
	    cn.cn_hash += *cp;

	cn.cn_cred = crget();
	cn.cn_cred->cr_uid = 0;

	cwd = curproc->p_fd->fd_cdir;
#if 0 /* XXX ? */
	VOP_LOCK(cwd, 0, p);
#endif 
	errno = VOP_LOOKUP(cwd, &vp, &cn);

	crfree(cn.cn_cred);

	if (errno == 0) {
	    if (DATA_FROM_XNODE(t))
		vrele(DATA_FROM_XNODE(t));
	    VOP_UNLOCK(vp, 0, p);
	    DATA_FROM_XNODE(t) = vp;
	    XFSDEB(XDEBMSG, ("xfs_message_installdata: t = %p ; tokens = %x\n",
			     t, message->node.tokens));

	    t->tokens = message->node.tokens;
	    xfs_attr2vattr(&message->node.attr, &t->attr);
	    bcopy(message->node.id, t->id, sizeof(t->id));
	    bcopy(message->node.rights, t->rights, sizeof(t->rights));
	} else
	    printf("XFS PANIC Warning: xfs_message_installdata failed to lookup cache file = %s, errno = %d\n", tmp, errno);
    } else {
	printf("XFS PANIC Warning: xfs_message_installdata didn't find node!\n");
	errno = ENOENT;
    }

    return errno;
}

int
xfs_message_invalidnode(int fd,
			struct xfs_message_invalidnode * message,
			u_int size)
{
    int errno = 0;
    struct xfs_node *t;

    XFSDEB(XDEBMSG, ("xfs_message_invalidnode\n"));

    t = xfs_node_find(&xfs[fd], &message->handle);
    if (t != 0) {
	/* XXX Really need to put back dirty data first. */
	if (DATA_FROM_XNODE(t)) {
	    vrele(DATA_FROM_XNODE(t));
	    DATA_FROM_XNODE(t) = (struct vnode *) 0;
	}
	XFS_TOKEN_CLEAR(t, ~0,
			XFS_OPEN_MASK | XFS_ATTR_MASK |
			XFS_DATA_MASK | XFS_LOCK_MASK);
	cache_purge(XNODE_TO_VNODE(t));
    } else {
	printf("XFS PANIC Warning: xfs_message_invalidnode "
	       "didn't find node!\n");
	errno = ENOENT;
    }

    return errno;
}
